"""
Return the CPU information for Solaris-like systems
    Provides:
      cpuarch
      num_cpus
      cpu_model
      cpu_flags
"""
import asyncio
import os
import re
import shutil


async def load_cpu_arch(hub):
    isainfo = shutil.which("isainfo")
    if isainfo:
        ret = await hub.exec.cmd.run([isainfo, "-k"])
        hub.corn.CORN.cpuarch = ret.stdout.strip()
    else:
        hub.corn.CORN.cpuarch = hub.corn.CORN.get("osarch")


async def load_cpu_flags(hub):
    cpu_flags = []

    isainfo = shutil.which("isainfo")

    if isainfo:
        ret = await hub.exec.cmd.run([isainfo, "-n", "-v"])
        for line in ret.stdout.splitlines():
            match = re.match(r"^\s+(.+)", line)
            if match:
                cpu_flags.extend(match.group(1).split())

    hub.corn.CORN.cpu_flags = sorted(set(cpu_flags))

    # Report if hardware virtualization is available under amd or intel
    hub.corn.CORN.hardware_virtualization = any(
        f in hub.corn.CORN.cpu_flags for f in ("svm", "vmx")
    )


async def load_cpu_model(hub):
    kstat = shutil.which("kstat")
    ret = await hub.exec.cmd.run([kstat, "-p", "cpu_info:*:*:brand"], shell=True)
    for line in ret.stdout.splitlines():
        match = re.match(r"(\w+:\d+:\w+\d+:\w+)\s+(.+)", line)
        if match:
            hub.corn.CORN.cpu_model = match.group(2)
            break


async def load_num_cpus(hub):
    psrinfo = shutil.which("psrinfo") or "/usr/sbin/psrinfo"
    if os.path.exists(psrinfo):
        ret = await hub.exec.cmd.run(psrinfo, stderr=asyncio.subprocess.DEVNULL)
        hub.corn.CORN.num_cpus = len(ret.stdout.strip().splitlines())



import getpass
import pwd


async def load_console_user(hub):
    hub.corn.CORN.console_username = getpass.getuser()
    hub.corn.CORN.console_user = pwd.getpwnam(hub.corn.CORN.console_username).pw_uid



import aiofiles
import logging
import os
import re
import shutil
from typing import Tuple

log = logging.getLogger(__name__)

_OS_FAMILY_MAP = {
    "Solaris": "Solaris",
    "SmartOS": "Solaris",
    "OmniOS": "Solaris",
    "OpenIndiana Development": "Solaris",
    "OpenIndiana": "Solaris",
    "OpenSolaris Development": "Solaris",
    "OpenSolaris": "Solaris",
    "Oracle Solaris": "Solaris",
}


async def load_osbuild(hub):
    hub.corn.CORN.osbuild = "unknown"

    pkg = shutil.which("pkg") or "/usr/bin/pkg"
    if os.path.exists(pkg):
        ret = await hub.exec.cmd.run([pkg, "info", "kernel"])
        hub.corn.CORN.osbuild = ret.stdout.strip().split("FMRI")[-1].split(":")[-1]


def _load_oscodename(hub, osname: str, osrelease: float) -> str:
    if ("sunos" in osname and osrelease <= 4.1) or (
        "solaris" in osname and osrelease <= 1.1
    ):
        return "Valkyrie"
    elif ("sunos" in osname and osrelease <= 5) or (
        "solaris" in osname and osrelease <= 2
    ):
        return "Jupiter"
    elif "solaris" in osname:
        if osrelease < 2.5:
            return "Starburst"
        elif osrelease <= 2.6:
            return "Wave3"
        elif osrelease == 7:
            return "StoreEdge N8200"
        elif "64" in hub.corn.CORN.get("osarch", ""):
            # For the full x86_64 version of Solaris
            return "Wyoming"
        else:
            # For the current release of Solaris
            return "Nevada"
    else:
        return "unknown"
    # Chrysalis Client if Solaris intel-based NC


async def load_manufacturer(hub):
    hub.corn.CORN.osmanufacturer = "unknown"
    prtconf = shutil.which("prtconf") or "/usr/sbin/prtconf"
    if os.path.exists(prtconf):
        ret = await hub.exec.cmd.run(prtconf)
        for line in ret.stdout.splitlines():
            if "System Configuration" in line:
                # remove the system configuration line and os arch
                hub.corn.CORN.osmanufacturer = " ".join(line.split()[2:-1])
                break


async def _load_release(hub) -> Tuple[str]:
    release_file = "/etc/release"
    if os.path.isfile(release_file):
        async with aiofiles.open(release_file, "r") as fp_:
            rel_data = await fp_.read()
            try:
                release_re = re.compile(
                    r"((?:(?:Open|Oracle )?Solaris|OpenIndiana|OmniOS)\s*(?:Development)?)"
                    r"\s*(\d+\.?\d*|v\d+)\s?[A-Z]*\s?(r\d+|\d+\/\d+|oi_\S+|snv_\S+)?"
                )
                return release_re.search(rel_data).groups()
            except AttributeError:
                pass


async def load_osinfo(hub):
    # Collect some preliminary information
    (osname, osmajorrelease, osminorrelease) = await _load_release(hub) or [
        "Solaris",
        "0",
        "0",
    ]

    # Load os
    hub.corn.CORN.os = "SmartOS" if hub.corn.CORN.smartos else osname.strip()
    hub.corn.CORN.os_family = _OS_FAMILY_MAP.get(hub.corn.CORN.os, hub.corn.CORN.os)

    # Sanitize the osrelease information
    if hub.corn.CORN.smartos:
        # See https://github.com/joyent/smartos-live/issues/224
        osrelease_stamp = hub.corn.CORN.kernelversion[
            hub.corn.CORN.kernelversion.index("_") + 1 :
        ]
        hub.corn.CORN.osrelease = ".".join(
            (x or "0")
            for x in (
                osrelease_stamp.split("T")[0][0:4],
                osrelease_stamp.split("T")[0][4:6],
                osrelease_stamp.split("T")[0][6:8],
            )
        )
    elif (
        hub.corn.CORN.os == "Oracle Solaris"
        and hub.corn.CORN.kernelversion.startswith(osmajorrelease)
    ):
        # Oracla Solars 11 and up have minor version in kernelversion
        hub.corn.CORN.osrelease = hub.corn.CORN.kernelversion[
            : hub.corn.CORN.kernelversion.find(".", 5)
        ]
    elif hub.corn.CORN.os == "OmniOS":
        hub.corn.CORN.osrelease = f"{osmajorrelease[1:] or 0}.{osminorrelease[1:] or 0}"
    else:
        hub.corn.CORN.osrelease = ""

    # Load release info
    hub.corn.CORN.osrelease_info = [
        int(x) if x.isdigit() else x or 0 for x in hub.corn.CORN.osrelease.split(".")
    ]
    assert len(hub.corn.CORN.osrelease_info)
    hub.corn.CORN.osmajorrelease = int(hub.corn.CORN.osrelease_info[0])
    hub.corn.CORN.osrelease = ".".join(
        str(x or 0) for x in hub.corn.CORN.osrelease_info
    )

    # Load other name info
    if len(hub.corn.CORN.osrelease_info) > 1:
        hub.corn.CORN.oscodename = _load_oscodename(
            hub,
            osname=hub.corn.CORN.os.lower(),
            osrelease=hub.corn.CORN.osrelease_info[0]
            + (hub.corn.CORN.osrelease_info[1] / 10),
        )
    else:
        hub.corn.CORN.oscodename = "unknown"

    if hub.corn.CORN.osmajorrelease:
        hub.corn.CORN.osfinger = f"{hub.corn.CORN.os}-{hub.corn.CORN.osmajorrelease}"
        # In salt, osfullname was a duplicate of what osfinger should have been and osfinger didn't exist
        # It's set straight now
        hub.corn.CORN.osfullname = f"{hub.corn.CORN.os}-{hub.corn.CORN.osrelease}"
    else:
        hub.corn.CORN.osfullname = hub.corn.CORN.osfinger = hub.corn.CORN.os



import os
import grp


async def load_group(hub):
    hub.corn.CORN.gid = os.getegid()
    try:
        hub.corn.CORN.groupname = grp.getgrgid(hub.corn.CORN.gid).gr_name
    except KeyError:
        hub.corn.CORN.groupname = None



import logging
import shutil
import socket
from typing import List

log = logging.getLogger(__name__)

# Possible value for h_errno defined in netdb.h
HOST_NOT_FOUND = 1
NO_DATA = 4


async def _get_fqdns(fqdn: str, protocol: int) -> List[str]:
    try:
        result = socket.getaddrinfo(
            host=fqdn,
            port=None,
            family=protocol,
            proto=socket.IPPROTO_IP,
            flags=socket.AI_NUMERICSERV | socket.AI_ADDRCONFIG | socket.AI_PASSIVE,
        )
        return sorted({item[4][0] for item in result})
    except socket.gaierror as e:
        log.debug(e)
    return []


async def load_socket_info(hub):
    hub.corn.CORN.localhost = socket.gethostname()

    hostname = shutil.which("hostname")
    if hostname:
        hub.corn.CORN.computer_name = (await hub.exec.cmd.run(hostname)).stdout.strip()
        hub.corn.CORN.fqdn = (await hub.exec.cmd.run([hostname, "-f"])).stdout.strip()
    else:
        hub.corn.CORN.computer_name = hub.corn.CORN.localhost

    # try socket.getaddrinfo to get fqdn
    try:
        addrinfo = socket.getaddrinfo(
            hub.corn.CORN.localhost,
            0,
            socket.AF_UNSPEC,
            socket.SOCK_STREAM,
            socket.SOL_TCP,
            socket.AI_CANONNAME,
        )
        for info in addrinfo:
            # info struct [family, socktype, proto, canonname, sockaddr]
            if len(info) >= 4:
                hub.corn.CORN.fqdn = info[3]

    except socket.gaierror:
        pass

    if not hub.corn.CORN.get("fqdn"):
        hub.corn.CORN.fqdn = socket.getfqdn()

    hub.corn.CORN.host, hub.corn.CORN.domain = hub.corn.CORN.fqdn.partition(".")[::2]

    hub.corn.CORN.fqdn_ip4 = await _get_fqdns(hub.corn.CORN.fqdn, socket.AF_INET)
    hub.corn.CORN.fqdn_ip6 = await _get_fqdns(hub.corn.CORN.fqdn, socket.AF_INET6)
    hub.corn.CORN.fqdns = hub.corn.CORN.fqdn_ip4 + hub.corn.CORN.fqdn_ip6



# -*- coding: utf-8 -*-
import asyncio
import functools
import logging
import os
import shlex
from typing import Any, Dict, List

log = logging.getLogger(__name__)

__virtualname__ = "cmd"


async def _sanitize_env(env: Dict[str, Any]) -> Dict[str, str] or None:
    if env is None:
        return
    for bad_env_key in (k for k, v in env.items() if v is None):
        log.error(
            "Environment variable '%s' passed without a value. "
            "Setting value to an empty string",
            bad_env_key,
        )
        env[bad_env_key] = ""
    return env


async def _sanitize_cwd(cwd: str or None) -> str:
    # salt-minion is running as. Defaults to home directory of user under which
    # the minion is running.
    if not cwd:
        cwd = os.path.expanduser("~")

        # make sure we can access the cwd
        # when run from sudo or another environment where the euid is
        # changed ~ will expand to the home of the original uid and
        # the euid might not have access to it. See issue #1844
        if not os.access(cwd, os.R_OK):
            cwd = "/"
    else:
        # Handle edge cases where numeric/other input is entered, and would be
        # yaml-ified into non-string types
        cwd = str(cwd)

    if not os.path.isabs(cwd) or not os.path.isdir(cwd):
        raise SystemError(
            f"Specified cwd '{cwd}' either not absolute or does not exist"
        )

    return cwd


async def _sanitize_cmd(cmd: str or List[str]) -> str or List[str]:
    if not isinstance(cmd, list):
        cmd = cmd.split()
    # Return stripped command string copies to improve logging.
    # Use shlex.quote to properly escape whitespace and special characters in strings passed to shells
    if isinstance(cmd, list):
        cmd = [shlex.quote(str(x).strip()) for x in cmd]
    else:
        cmd = shlex.quote(cmd)
    return cmd


async def _sanitize_umask(umask: str) -> int or None:
    if umask is None:
        return

    _umask = str(umask).lstrip("0")

    if _umask == "":
        raise SystemError("Zero umask is not allowed.")

    try:
        return int(_umask, 8)
    except ValueError:
        raise SystemError("Invalid umask: '{0}'".format(umask))


async def _sanitize_kwargs(**kwargs):
    """
    Only pass through approved kwargs
    """
    new_kwargs = {}
    if "stdin_raw_newlines" in kwargs:
        new_kwargs["stdin_raw_newlines"] = kwargs["stdin_raw_newlines"]
    return new_kwargs


async def run(
    hub,
    cmd: str or List[str],
    cwd: str = None,
    shell: bool = False,
    stdin: str = None,
    stdout: int = asyncio.subprocess.PIPE,
    stderr: int = asyncio.subprocess.PIPE,
    env: Dict[str, Any] = None,
    umask: str = None,
    timeout: int or float = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    Execute the passed command and return the output as a string

    :param cmd: The command to run. ex: ``ls -lart /home``

    :param cwd: The directory from which to execute the command. Defaults
        to the home directory of the user specified by ``runas`` (or the user
        under which Salt is running if ``runas`` is not specified).

    :param stdin: A string of standard input can be specified for the
        command to be run using the ``stdin`` parameter. This can be useful in
        cases where sensitive information must be read from standard input.

    :param shell: If ``False``, let python handle the positional
        arguments. Set to ``True`` to use shell features, such as pipes or
        redirection.

    :param stdout:

    :param stderr:

    :param env: Environment variables to be set prior to execution.

        .. note::
            When passing environment variables on the CLI, they should be
            passed as the string representation of a dictionary.

            .. code-block:: bash

                salt myminion cmd.run 'some command' env='{"FOO": "bar"}'
    :param umask: The umask (in octal) to use when running the command.

    :param timeout: A timeout in seconds for the executed process to return.
    """
    cmd = await _sanitize_cmd(cmd)
    cwd = await _sanitize_cwd(cwd)
    env = await _sanitize_env(env)
    umask = await _sanitize_umask(umask)

    new_kwargs = {
        "cwd": cwd,
        "env": env if env else os.environ.copy(),
        "preexec_fn": functools.partial(os.umask, umask) if umask else None,
        "stdout": stdout,
        "stderr": stderr,
        "shell": shell,
        **await _sanitize_kwargs(**kwargs),
    }

    # Run the command
    if shell:
        proc = await asyncio.create_subprocess_shell(" ".join(cmd), **new_kwargs)
    else:
        proc = await asyncio.create_subprocess_exec(*cmd, **new_kwargs)

    # This is where the magic happens
    out, err = await asyncio.wait_for(proc.communicate(input=stdin), timeout=timeout)

    return hub.pop.data.imap(
        {
            "pid": proc.pid,
            "retcode": await asyncio.wait_for(proc.wait(), timeout=timeout),
            "stdout": (out or b"").decode(),
            "stderr": (err or b"").decode(),
        }
    )



"""
Return the memory information for SunOS-like systems
"""
import asyncio
import os
import shutil


async def load_meminfo(hub):
    hub.corn.CORN.mem_total = 0
    prtconf = shutil.which("prtconf") or "/usr/sbin/prtconf"

    if os.path.exists(prtconf):
        ret = await hub.exec.cmd.run(prtconf, stderr=asyncio.subprocess.DEVNULL)
        for line in ret.stdout.splitlines():
            comps = line.split(" ")
            if comps[0].strip() == "Memory" and comps[1].strip() == "size:":
                hub.corn.CORN.mem_total = int(comps[2].strip())
                break


async def load_swap(hub):
    hub.corn.CORN.swap_total = 0

    swap_cmd = shutil.which("swap")
    if swap_cmd:
        # total: 605092k bytes allocated + 401380k reserved = 1006472k used, 4116500k available
        ret = await hub.exec.cmd.run([swap_cmd, "-s"])
        swap_data = ret.stdout.strip().split()
        try:
            swap_avail = int(swap_data[-2][:-1])
            swap_used = int(swap_data[-4][:-1])
            hub.corn.CORN.swap_total = (swap_avail + swap_used) // 1024
        except ValueError:
            pass



import asyncio
import logging
import re
import shutil
from typing import Any, Dict, List, Tuple

log = logging.getLogger(__name__)


async def _cidr_to_ipv4_netmask(cidr_bits: int) -> str:
    """
    Returns an IPv4 netmask
    """
    try:
        cidr_bits = cidr_bits
        if not 1 <= cidr_bits <= 32:
            return ""
    except ValueError:
        return ""

    netmask = ""
    for idx in range(4):
        if idx:
            netmask += "."
        if cidr_bits >= 8:
            netmask += "255"
            cidr_bits -= 8
        else:
            netmask += "{0:d}".format(256 - (2 ** (8 - cidr_bits)))
            cidr_bits = 0
    return netmask


async def _number_of_set_bits(x):
    """
    Returns the number of bits that are set in a 32bit int
    """
    # Taken from http://stackoverflow.com/a/4912729. Many thanks!
    x -= (x >> 1) & 0x55555555
    x = ((x >> 2) & 0x33333333) + (x & 0x33333333)
    x = ((x >> 4) + x) & 0x0F0F0F0F
    x += x >> 8
    x += x >> 16
    return x & 0x0000003F


async def _number_of_set_bits_to_ipv4_netmask(set_bits: int) -> str:
    """
    Returns an IPv4 netmask from the integer representation of that mask.

    Ex. 0xffffff00 -> '255.255.255.0'
    """
    return await _cidr_to_ipv4_netmask(await _number_of_set_bits(set_bits))


async def _interfaces_ifconfig(out: str) -> Dict[str, Any]:
    """
    Uses ifconfig to return a dictionary of interfaces with various information
    about each (up/down state, ip address, netmask, and hwaddr)
    """
    ret = {}

    piface = re.compile(r"^([^\s:]+)")
    pip = re.compile(r".*?(?:inet addr:|inet [^\d]*)(.*?)\s")
    pip6 = re.compile(".*?(?:inet6 addr: (.*?)/|inet6 )([0-9a-fA-F:]+)")
    pmask6 = re.compile(r".*?(?:inet6 addr: [0-9a-fA-F:]+/(\d+)|)")
    pmask = re.compile(r".*?(?:Mask:|netmask )(?:((?:0x)?[0-9a-fA-F]{8})|([\d\.]+))")
    pupdown = re.compile("UP")
    pbcast = re.compile(r".*?(?:Bcast:|broadcast )([\d\.]+)")

    groups = re.compile("\r?\n(?=\\S)").split(out)
    for group in groups:
        data = {}
        iface = ""
        updown = False
        for line in group.splitlines():
            miface = piface.match(line)
            mip = pip.match(line)
            mip6 = pip6.match(line)
            mupdown = pupdown.search(line)
            if miface:
                iface = miface.group(1)
            if mip:
                if "inet" not in data:
                    data["inet"] = list()
                addr_obj = dict()
                addr_obj["address"] = mip.group(1)
                mmask = pmask.match(line)
                if mmask:
                    if mmask.group(1):
                        mmask = await _number_of_set_bits_to_ipv4_netmask(
                            int(mmask.group(1), 16)
                        )
                    else:
                        mmask = mmask.group(2)
                    addr_obj["netmask"] = mmask
                mbcast = pbcast.match(line)
                if mbcast:
                    addr_obj["broadcast"] = mbcast.group(1)
                data["inet"].append(addr_obj)
            if mupdown:
                updown = True
            if mip6:
                if "inet6" not in data:
                    data["inet6"] = list()
                addr_obj = dict()
                addr_obj["address"] = mip6.group(1) or mip6.group(2)
                mmask6 = pmask6.match(line)
                if mmask6:
                    if addr_obj["address"] != "::":
                        data["inet6"].append(addr_obj)
        data["up"] = updown
        if iface in ret:
            # SunOS optimization, where interfaces occur twice in 'ifconfig -a'
            # output with the same name: for ipv4 and then for ipv6 addr family.
            # Every instance has it's own 'UP' status and we assume that ipv4
            # status determines global interface status.
            #
            # merge items with higher priority for older values
            # after that merge the inet and inet6 sub items for both
            ret[iface].update(data)
            if "inet" in data:
                ret[iface]["inet"].extend(
                    x for x in data["inet"] if x not in ret[iface]["inet"]
                )
            if "inet6" in data:
                ret[iface]["inet6"].extend(
                    x for x in data["inet6"] if x not in ret[iface]["inet6"]
                )
        else:
            ret[iface] = data
        del data
    return ret


async def _get_hwaddr_interfaces4(hub) -> Dict[str, str]:
    """
    Read netstat and look for mac addresses
    """
    result = {}
    netstat = shutil.which("netstat")
    if netstat:
        hwaddr = re.compile(
            "\w+\s+([\d|\.]+)\s+[\w|\.]+\s+\w*\s+((?:[0-9A-Fa-f]{2}[:-]){5}(?:[0-9A-Fa-f]{2}))$"
        )
        ret = await hub.exec.cmd.run([netstat, "-p", "-n", "-f", "inet"])
        for line in ret.stdout.splitlines():
            match = hwaddr.match(line.strip())
            if match:
                result[match.group(1)] = match.group(2)
    return result


async def _get_hwaddr_interfaces6(hub) -> Dict[str, str]:
    """
    Read netstat and look for mac addresses
    """
    result = {}
    netstat = shutil.which("netstat")
    if netstat:
        hwaddr = re.compile(
            "\w+\s+((?:[0-9A-Fa-f]{2}[:-]){5}(?:[0-9A-Fa-f]{2}))\s+\w+\s+\w+\s+([0-9a-fA-F|:]+)$"
        )
        ret = await hub.exec.cmd.run([netstat, "-p", "-n", "-f", "inet6"])
        for line in ret.stdout.splitlines():
            match = hwaddr.match(line.strip())
            if match:
                result[match.group(2)] = match.group(1)
    return result


async def load_interfaces(hub):
    """
    Provide a dict of the connected interfaces and their ip addresses
    The addresses will be passed as a list for each interface
    """

    log.debug("Loading interfaces")
    # Provides:
    #   ip_interfaces
    ipv4 = []
    ipv6 = []

    # Start collecting mac addresses
    hwaddrs = await _get_hwaddr_interfaces4(hub)
    hwaddrs.update(await _get_hwaddr_interfaces6(hub))

    ifconfig_path = shutil.which("ifconfig")
    if ifconfig_path:
        ret = await hub.exec.cmd.run([ifconfig_path, "-a"])
        interfaces = await _interfaces_ifconfig(str(ret.stdout.strip()))

        for interface, device in interfaces.items():
            if not interface:
                continue
            inet4: List[str] = [ip.get("address") for ip in device.get("inet", [])]
            ipv4.extend(inet4)
            if inet4:
                hub.corn.CORN.ip4_interfaces[interface] = sorted(inet4)
            inet6: List[str] = [ip.get("address") for ip in device.get("inet6", [])]
            ipv6.extend(inet6)
            if inet6:
                hub.corn.CORN.ip6_interfaces[interface] = sorted(inet6)
            hub.corn.CORN.ip_interfaces[interface] = sorted(inet4) + sorted(inet6)

            for addr in ipv4 + ipv6:
                hw_addr = hwaddrs.get(addr)
                if hw_addr:
                    hub.corn.CORN.hwaddr_interfaces[interface] = hw_addr
                    break
    hub.corn.CORN.ipv4 = sorted(ipv4)
    hub.corn.CORN.ipv6 = sorted(ipv6)



import os
import sys


async def load_uname(hub):
    """
    Verify that idem-solaris is running on SunOS
    """
    (
        hub.corn.CORN.kernel,
        hub.corn.CORN.nodename,
        hub.corn.CORN.kernelrelease,
        hub.corn.CORN.kernelversion,
        hub.corn.CORN.osarch,
    ) = os.uname()

    assert (
        hub.corn.CORN.kernel == "SunOS"
    ), "idem-solaris is only intended for SunOs systems"

    # Determine if host is SmartOS (Illumos) or not
    hub.corn.CORN.smartos = sys.platform.startswith(
        "sunos"
    ) and hub.corn.CORN.kernelversion.startswith("joyent_")

    # Hard coded grains for SunOs go here:
